// db_module.js : Server-side bindings for DB Display module
// Author: Eric Flumerfelt, FNAL RSI
// Last Modified: October 30, 2015
//

// Node.js framework "includes"
// ReSharper disable PossiblyUnassignedProperty
// ReSharper disable InconsistentNaming
var Emitter = require("events").EventEmitter;
var fs = require("fs");
var stream = require("stream");
var child_process = require("child_process");
var Utils = require("./Utils");
var Connector = require("./DBConnector");
var path_module = require("path");
var db = new Emitter();
// ReSharper restore InconsistentNaming
// ReSharper restore PossiblyUnassignedProperty

var DefaultColumns = [
    {
        name: "name",
        type: "string",
        editable: false,
        display: true
    },
    {
        name: "value",
        type: "string",
        editable: true,
        display: true
    },
    {
        name: "annotation",
        title: "User Comment",
        type: "string",
        editable: true,
        display: true
    },
    {
        name: "comment",
        type: "comment",
        editable: false,
        display: false
    },
    {
        name: "type",
        type: "string",
        editable: false,
        display: false
    },
    {
        name: "values",
        type: "array",
        editable: false,
        display: false
    }, {
        name: "children",
        type: "array",
        editable: false,
        display: false
    }
];

var MakeDbConfig = function (configFile) {
    var defaultConfig = {
        dbprovider: "filesystem"
        , configNameFilter: "*"
        , baseDir: process.env.ARTDAQ_DATABASE_URI ? process.env["ARTDAQ_DATABASE_URI"] : ""
        , mongoConfig: {
            dbUser: "anonymous"
            , dbHost: "localhost"
            , dbPort: 27017
        }
        , instanceName: configFile.instanceName ? configFile.instanceName : "test_db"
    };

    if (defaultConfig.baseDir.indexOf("filesystemdb://") === 0) {
        console.log("Parsing filesystemdb URI: " + defaultConfig.baseDir);
        defaultConfig.baseDir = defaultConfig.baseDir.replace("filesystemdb://", "").replace(/\/$/, "");
        var tempArr = defaultConfig.baseDir.split('/');
        defaultConfig.instanceName = tempArr.splice(-1);
        defaultConfig.baseDir = tempArr.slice(0,-1).join('/');
    } else if (defaultConfig.baseDir.indexOf("mongodb://") === 0) {
        console.log("Parsing mongodb URI: " + defaultConfig.baseDir);
        var matches = defaultConfig.baseDir.match(/mongodb:\/\/(?:([^@:]*)(?::([^@]*))?@)?([^:]*):([0-9]+)\/([^/]*)\/?/);
        defaultConfig.baseDir = "";
        defaultConfig.dbprovider = "mongo";
        console.log(JSON.stringify(matches, null, 4));
        defaultConfig.mongoConfig.dbUser = matches[1] ? matches[1] : "";
        var pass = matches[2];
        if (pass !== null && pass !== undefined) {
            fs.writeFileSync(".dbpasswd", pass);
        }
        defaultConfig.mongoConfig.dbHost = matches[3];
        defaultConfig.mongoConfig.dbPort = matches[4];
        defaultConfig.instanceName = matches[5];
    }

    if (defaultConfig.baseDir === "") {
        defaultConfig.baseDir = process.env["ARTDAQ_DATABASE_DIR"];
    }

    console.log("DBConfig: " + JSON.stringify(defaultConfig, null, 4));
    return defaultConfig;
};

function GetFileBase(configName, collection, entity) {
    return path_module.join(configName, collection, entity + ".gui.json");
}

function GetTempFilePath(configName, collection, entity, dirs) {
    var filebase = GetFileBase(configName, collection, entity);

    if (!fs.existsSync(path_module.join(dirs.tmp, configName))) fs.mkdirSync(path_module.join(dirs.tmp, configName));
    if (!fs.existsSync(path_module.join(dirs.tmp, configName, collection))) fs.mkdirSync(path_module.join(dirs.tmp, configName, collection));
    return path_module.join(dirs.tmp, filebase);
}

function GetFilePath(configName, collection, entity, dirs, dbOnly) {
    var filebase = GetFileBase(configName, collection, entity);
    console.log("GetFilePath: filebase is " + filebase + ", dirs: " + JSON.stringify(dirs, null, 4));
    if (!dbOnly && fs.existsSync(path_module.join(dirs.tmp, filebase))) {
        return path_module.join(dirs.tmp, filebase);
    }

    if (!fs.existsSync(path_module.join(dirs.db, configName))) fs.mkdirSync(path_module.join(dirs.db, configName));
    if (!fs.existsSync(path_module.join(dirs.db, configName, collection))) fs.mkdirSync(path_module.join(dirs.db, configName, collection));
    return path_module.join(dirs.db, filebase);
}

/**
 * Loads a file from the database and saves it to disk, if it doesn't already exist
 * @param fileInfo - fileInfo.query should be the query needed to load the file
 * @param dbDirectory - Directory to save the file into
 */
function LoadFile(fileInfo, dirs, dbConfig) {
    console.log("LoadFile: fileInfo: " + JSON.stringify(fileInfo, null, 4) + ", dirs: " + JSON.stringify(dirs, null, 4) + ", dbConfig: " + JSON.stringify(dbConfig, null, 4));
    var output = {
        data: {}
        , filePath: ""
        , entity: ""
        , collection: ""
        , configName: ""
    };

    output.entity = fileInfo.query.filter["entities.name"];
    output.collection = fileInfo.query.collection;
    output.configName = fileInfo.query.filter["configurations.name"];

    console.log("File info: " + JSON.stringify(fileInfo, null, 4));
    output.filePath = GetFilePath(output.configName, output.collection, output.entity, dirs, true);

    if (!fs.existsSync(output.filePath)) {
        console.log("Loading file from database");
        output.data = Connector.RunLoadQuery(dbConfig, fileInfo.query);
        console.log("Writing file " + output.filePath);
        fs.writeFileSync(output.filePath, JSON.stringify(output.data, null, 4));
    } else {
        output.data = JSON.parse("" + fs.readFileSync(output.filePath));
    }
    console.log("File Data: " + JSON.stringify(output, null, 4));
    console.log("Done Loading file");
    return output;
}

/**
 * Reads a file out of the database and returns information on where to find it
 * @param {Object} fileInfo - Input search criteria
 * @param {string} fileInfo.name - Configurable Entity Name
 * @param {string} fileInfo.collection - Collection Name
 * @param {string} fileInfo.version - Configuration Version
 * @param {string} dataFormat - Data format (fhicl, json, gui)
 * @param {string} dbdirectory - Database storage directory
 * @returns {Object} Information abouit the output: fileName, filePath, and size
 * @throws DBOperationFailedException: More information in ex.message
 */
function FetchFile(fileInfo, dataFormat, dbdirectory, dbConfig) {
    console.log("FetchFile: fileInfo: " + JSON.stringify(fileInfo, null, 4) + ", dataFormat: " + dataFormat + ", dbDirectory: " + dbdirectory);
    var fileName = path_module.join(fileInfo.collection + "_" + fileInfo.name + "_" + fileInfo.version);
    if (dataFormat === "fhicl") {
        fileName += ".fcl";
    } else if (dataFormat === "gui") {
        fileName += ".gui.json";
    } else {
        fileName += ".json";
    }
    var filePath = path_module.join(dbdirectory, fileName);
    var query = {
        filter: {
            "entities.name": fileInfo.name,
            version: fileInfo.version
        },
        collection: fileInfo.collection,
        configurable_entity: fileInfo.name,
        dbprovider: dbConfig.dbprovider,
        operation: "load",
        dataformat: dataFormat
    };
    var fhiclData = Connector.RunLoadQuery(dbConfig, query, true);
    fs.writeFileSync(filePath, JSON.stringify(fhiclData, null, 4));

    var stat = fs.statSync(filePath);
    return { fileName: fileName, filePath: filePath, size: stat.size }
};

function MakeColumns(rowIn, index, columns, columnGroups, group) {
    //console.log("MakeColumns: rowIn: " + JSON.stringify(rowIn,null,4) + ", index: " + index + ", columns: " + JSON.stringify(columns,null,4) + ", columnGroups: " + JSON.stringify(columnGroups,null,4) + ", group: " + group);
    var rowOut = {
        children: []
    };

    for (var p in rowIn) {
        if (rowIn.hasOwnProperty(p)) {
            var value = rowIn[p];
            if (value !== Object(value) && Utils.ContainsName(columns, p, "name") < 0) {
                columns.push({ name: p, type: "string", editable: true, display: true, columnGroup: group });
                rowOut[p] = value;
            } else if (value === Object(value)) {
                var res;
                if (Utils.ContainsName(columnGroups, p, "name") < 0) {
                    columnGroups.push({ name: "" + p, text: "" + p, parentGroup: group });
                }
                if (value.constructor === Array) {
                    for (var i in value) {
                        if (value.hasOwnProperty(i)) {
                            res = MakeColumns(value[i], index + "." + i, columns, columnGroups, p);
                            columns = res.columns;
                            columnGroups = res.columnGroups;
                            res.row.name = rowIn.name + "/" + p + "___" + i;
                            res.row.index = index + "." + i;
                            rowOut.children.push(res.row);
                        }
                    }
                } else {
                    res = MakeColumns(value, index, columns, columnGroups, p);
                    columns = res.columns;
                    columnGroups = res.columnGroups;
                    for (var pp in res.row) {
                        if (res.row.hasOwnProperty(pp)) {
                            rowOut[pp] = res.row[pp];
                        }
                    }

                }
            } else {
                rowOut[p] = value;
            }
        }
    }
    if (rowOut.children.length === 0) delete rowOut.children;
    else if (Utils.ContainsName(columns, "children", "name") < 0) { columns.push({ name: "children", type: "array", editable: false, display: false }); }

    return { row: rowOut, columns: columns, columnGroups: columnGroups };
}

/**
 * Transforms FHiCL sequences into arrays of objects for display by GUI editor
 * @param {Object} sequence - Sequence array
 * @param {string} name - Name of the Sequence
 * @returns {Object} Sqeuence object after transform
 */
function ParseSequence(sequence, name) {
    console.log("ParseSequence: sequence: " + JSON.stringify(sequence, null, 4) + ", name: " + name);
    var children = [];
    var hasTable = false;
    var rows = [];
    var columns = [];
    var columnGroups = [];
    columns.push({ name: "index", type: "number", editable: false, display: true, columnGroup: name });
    columns.push({ name: "name", type: "string", editable: false, display: true, columnGroup: name });
    //console.log("SEQUENCE BEFORE: " + JSON.stringify(sequence,null,4));
    for (var i = 0; i < sequence.length; ++i) {
        if (sequence[i] === Object(sequence[i])) {
            // Is an Object or Array
            if (sequence[i].constructor === Array) {
                // Is an array
                var arr = ParseSequence(sequence[i], name + "___" + i);
                if (arr.children !== undefined && arr.children.length > 0) {
                    children.push({ name: name + "___" + i, children: arr.children });
                    rows.push({ index: i, name: name + "___" + i, children: arr.children });
                    if (Utils.ContainsName(columns, "children", "name") < 0) {
                        columns.push({ name: "children", type: "array", editable: false, display: false });
                    }
                }
                if (arr.rows !== undefined && arr.rows.length > 0) {
                    if (arr.hasTable) {
                        hasTable = true;
                    }
                    for (var r in arr.rows) {
                        if (arr.rows.hasOwnProperty(r)) {
                            rows.push(arr.rows[r]);
                        }
                    }
                    for (var c in arr.columns) {
                        if (arr.columns.hasOwnProperty(c) && Utils.ContainsName(columns, arr.columns[c].name, "name") < 0) {
                            columns.push(arr.columns[c]);
                        }
                    }
                }
            } else {
                // We need to make a new table
                hasTable = true;
                console.log("Parsing as table: " + JSON.stringify(sequence[i], null, 4));
                if (!sequence[i].hasOwnProperty("name") || sequence[i].name.length === 0) {
                    sequence[i].name = name + "___" + i;
                }
                if (Utils.ContainsName(columnGroups, name, "name") < 0) {
                    columnGroups.push({ name: "" + name, text: "" + name });
                }
                var res = MakeColumns(sequence[i], i, columns, columnGroups, name);
                res.row.index = i;
                rows.push(res.row);
                columns = res.columns;
                columnGroups = res.columnGroups;
            }
        } else {
            var vname = name + "___" + i;
            var value = sequence[i];
            children.push({ name: vname, value: value });
            rows.push({ index: i, name: vname, value: value });
            if (Utils.ContainsName(columns, "value", "name") < 0) {
                columns.push({ name: "value", type: "string", editable: true, display: true, columnGroup: name });
            }
        }
    }
    var comment = sequence.comment ? sequence.comment : " ";

    if (hasTable) {
        var table = {
            name: name
            , comment: comment
            , type: "table"
            , isSequence: true
            , table: {
                rows: rows
                , columns: columns
                , columnGroups: columnGroups
            }
        };
        console.log("\nParseSequence returning table: " + JSON.stringify(table, null, 4));
        return table;
    } else {
        var output = { name: name, comment: comment, children: children };
        console.log("\nParseSequence returning: " + JSON.stringify(output, null, 4));
        return output;
    }
};

/**
 * Fills the "hasSubtables" variable for each table object. Creates "Table Entries" tables so that no table contains both subtables and values.
 * @param {Object} table - FHiCL table data to parse
 * @param {Number} sub - Dummy variable to limit debug output to final return statement only
 * @returns {Object} Converted FHiCL data
 */
function ParseFhiclTable(table, sub) {
    console.log("ParseFhiclTable: table: " + JSON.stringify(table, null, 4) + ", sub: " + sub);
    var children = [];
    var subtables = [];
    var hasSubtables = false;
    var comment = table.comment ? table.comment : " ";
    //console.log("Table name is " + name);

    for (var e in table.children) {
        if (table.children.hasOwnProperty(e)) {
            var element = table.children[e];
            if (element.type === "table" && element.isSequence) {
                element.type = "sequence";
            }
            //console.log("Element: " + JSON.stringify(element,null,4));
            switch (element.type) {
                case "table":
                    //console.log("Parsing table " + e);
                    hasSubtables = true;
                    children.push(ParseFhiclTable(element, 1));
                    break;
                case "sequence":
                    children.push(ParseSequence(element.children, element.name));
                    break;
                case "number":
                case "string":
                case "bool":
                    children.push(element);
                    break;
                default:
                    console.log("Unknown type " + element.type + " encountered!");
                    break;
            }
        }
    }

    var obj = { name: table.name, hasSubtables: hasSubtables, children: children, subtables: subtables, type: "table", comment: comment, columns: DefaultColumns };
    if (sub === 0 || sub === undefined) {
        //console.log("Returning: " + JSON.stringify(obj,null,4));
    }
    return obj;
};

function FindConfigs(configNameFilter, dbConfig) {
    console.log("\nFindConfigs: Request for Named Configurations received");
    var configsOutput = { Success: false, data: {}, configs: {} };
    try {
        var configs = Connector.RunGetConfigsQuery(dbConfig).search;
        console.log("CONFIGS: " + JSON.stringify(configs));
        for (var conf in configs) {
            if (configs.hasOwnProperty(conf)) {
                var config = configs[conf];
                if (config.name.search(configNameFilter) >= 0) {
                    var prefix = config.name;
                    var configNumber = 0;
                    var matches = config.name.match(/(.*?)[_\-]*(\d*)$/);
                    if (matches) {
                        if(matches.length > 1) prefix = matches[1];
                        if(matches.length > 2) configNumber = parseInt(matches[2]);
                    }

                    if (!configsOutput.data.hasOwnProperty(prefix)) {
                        configsOutput.data[prefix] = [];
                    }

                    configsOutput.data[prefix].push({ version: configNumber, data: JSON.stringify(config.query), name: config.name });
                    configsOutput.configs[config.name] = config.query;
                }
            }
        }
        if (configsOutput.data.length === 0) {
            configsOutput.data["null"] = {version: 0, data: ""};
            configsOutput.configs[configNameFilter] = {};
        }
        configsOutput.Success = true;
    } catch (e) {
        console.log("Exception caught: " + e.name + ": " + e.message);
    }
    console.log("NamedConfigs complete");
    return configsOutput;
}

function FindInstances(dbConfig) {
    console.log("FindInstances: Request for Database Instances received");
    var instancesOutput = { Success: false, data: [] };
    instancesOutput.data.push("<option value=\"NONE\">Create New Database Instance</option>");
    try {
        var dbs = Connector.RunListDatabasesQuery(dbConfig, {}).search;
        for (var db in dbs) {
            if (dbs.hasOwnProperty(db)) {
                instancesOutput.data.push("<option value=" + dbs[db].name + ">" + dbs[db].name + "</option>");
            }
        }
        instancesOutput.Success = true;
    } catch (e) {
        console.log("Exception caught: " + e.name + ": " + e.message);
    }
    console.log("FindInstances complete");
    return instancesOutput;
}

/**
 * Returns the list of files present in a configuration, after loading them to temporary storage
 * @param {string} configName - Name of the configuration to load
 * @param {string} dbDirectory - The database temporary storage directory
 * @param {Object} query - Database query for buildFilter
 * @returns {Object} {}.files contains the names of the files, .Success is whether the operation succeeded.
 */
function LoadConfigFiles(configName, dirs, query, dbConfig) {
    console.log("LoadConfigFiles: configName: " + configName + ", dirs: " + JSON.stringify(dirs, null, 4) + ", query: " + JSON.stringify(query, null, 4));
    var retval = {
        collections: {},
        Success: false
    };
    var error = false;
    var configFiles = [];
    var e;
    try {
        configFiles = Connector.RunBuildFilterQuery(dbConfig, configName, query).search;
    } catch (e) {
        error = true;
        console.log("Exception occurred: " + e.name + ": " + e.message);
    }
    if (!error) {
        try {
            for (var file in configFiles) {
                if (configFiles.hasOwnProperty(file)) {
                    console.log("File info: " + JSON.stringify(configFiles[file], null, 4));
                    var entity = LoadFile(configFiles[file], dirs, dbConfig).entity;
                    var collection = configFiles[file].query.collection;
                    if (!retval.collections.hasOwnProperty(collection)) {
                        retval.collections[collection] = {
                            name: collection
                            , files: []
                        };
                    }
                    console.log("Adding " + entity + " to output list");
                    retval.collections[collection].files.push(entity);
                }
            }
            retval.Success = true;
        } catch (e) {
            console.log("Exception occurred: " + e.name + ": " + e.message);
        }
    }
    return retval;
};

function SearchFile(fileData, searchKey) {
    console.log("\nSearching for " + searchKey + " in object " + JSON.stringify(fileData, null, 4));
    var retval = { name: fileData.name, children: [], columns: fileData.columns };
    for (var ii in fileData.children) {
        if (fileData.children.hasOwnProperty(ii)) {
            var child = fileData.children[ii];
            console.log("Key name: " + child.name);
            if (child.name.indexOf(searchKey) !== -1) {
                console.log("Adding child to output");
                retval.children.push(child);
            } else if (child.children && child.children.length > 0) {
                console.log("Searching child's children");
                var table = SearchFile(child, searchKey);
                if (table.children.length > 0) {
                    retval.children.push(table);
                }
            }
        }
    }
    console.log("\nSearchFile returning " + JSON.stringify(retval, null, 4));
    return retval;
}

function SearchConfigFiles(searchKey, configName, dirs, dbConfig) {
    console.log("\nSearchConfigFiles: keySearch: " + searchKey + ", configName: " + configName + ", dirs: " + JSON.stringify(dirs, null, 4));
    var retval = {
        collectionsTemp: {},
        collections: [],
        columns: DefaultColumns,
        Success: false
    };
    var error = false;
    var configFiles = [];
    var query = FindConfigs(configName, dbConfig).configs[configName];
    var e;
    try {
        configFiles = Connector.RunBuildFilterQuery(dbConfig, configName, query).search;
    } catch (e) {
        error = true;
        console.log("Exception occurred: " + e.name + ": " + e.message);
    }
    if (!error) {
        try {
            for (var file in configFiles) {
                if (configFiles.hasOwnProperty(file)) {
                    console.log("File info: " + JSON.stringify(configFiles[file], null, 4));
                    var entity = LoadFile(configFiles[file], dirs, dbConfig).entity;
                    var collection = configFiles[file].query.collection;
                    if (!retval.collectionsTemp.hasOwnProperty(collection)) {
                        retval.collectionsTemp[collection] = {
                            name: collection
                            , children: []
                        };
                    }
                    var fileData = GetData(configName, collection, entity, dirs);
                    fileData.name = entity;
                    var fileMatches = SearchFile(fileData, searchKey);
                    if (fileMatches.children.length > 0) {
                        console.log("Adding " + entity + " to output list");
                        retval.collectionsTemp[collection].children.push(fileMatches);
                    }
                }
            }
            retval.Success = true;
        } catch (e) {
            console.log("Exception occurred: " + e.name + ": " + e.message);
        }
    }

    for (var ii in retval.collectionsTemp) {
        if (retval.collectionsTemp.hasOwnProperty(ii)) {
            retval.collections.push(retval.collectionsTemp[ii]);
        }
    }
    delete retval.collectionsTemp;
    console.log("\nSearchConfigFiles returning: " + JSON.stringify(retval, null, 4));
    return retval;
}

/**
 * Returns a table from a configuration
 * @param {string} configName - Name of the configuration
 * @param {string} collectionName - Name of the collection
 * @param {string} filebase - Name of the configurable entity
 * @param {Object} dirs - Object containing paths to important directories
 * @param {string} dirs.tmp - The temporary directory for this editing session
 * @param {string} dirs.db - The database temporary storage directory
 * @returns {Object} Table Object
 * @throws FileNotFoundException - The requested file was not found
 * @throws TableNotFoundException - The named table was not found
 */
function GetData(configName, collectionName, entity, dirs) {
    console.log("GetData: configName: " + configName + ", collectionName: " + collectionName + ", entity: " + entity + ", dirs: " + JSON.stringify(dirs, null, 4));
    var fileName = GetFilePath(configName, collectionName, entity, dirs, false);
    if (!fs.existsSync(fileName)) { throw { name: "FileNotFoundException", message: "The requested file was not found" }; }
    var jsonFile = JSON.parse("" + fs.readFileSync(fileName));
    var jsonBase = ParseFhiclTable({ children: jsonFile.document.converted.guidata, name: entity }, 0);

    return jsonBase;
};

function ReplaceByPath(obj, pathArr, data) {
    console.log("Utils.ReplaceByPath: obj: " + JSON.stringify(obj, null, 4) + ", pathArr: " + JSON.stringify(pathArr, null, 4) + ", data: " + JSON.stringify(data, null, 4));
    if (pathArr.length === 0 && (!obj.name || obj.name === data.name)) {
        if (data.column && data.value) {
            obj[data.column] = data.value;
        } else if (data.type) {
            for (var i in data) {
                if (data.hasOwnProperty(i)) {
                    obj[i] = data[i];
                }
            }
        }
        return obj;
    }

    var thisName = pathArr.shift();
    var index = -1;
    if (obj.type === "table") {
        index = Utils.ContainsName(obj.children, thisName, "name");
    } else if (obj.type === "sequence") {
        index = thisName.slice(thisName.indexOf("___") + 3);
    }
    //console.log("Index is " + index + " (thisName: " + thisName + ", obj.children: " + JSON.stringify(obj.children,null,4) + ")");
    obj.children[index] = ReplaceByPath(obj.children[index], pathArr, data);
    return obj;
}

/**
 * Updates a value in a table and calles SetTable to save changes to disk
 * @param {string} configPath - Name of the Configuration
 * @param {string} tablePath - Path to the table, unix style (Config File Name/path/to/table)
 * @param {string} data - Modified value data
 * @param {Object} dirs - Object containing paths to important directories
 * @param {string} dirs.tmp - The temporary directory for this editing session
 * @param {string} dirs.db - The database temporary storage directory
 */
function UpdateTable(configName, tablePath, data, dirs) {
    console.log("UpdateTable: tablePath:" + tablePath + ", configName: " + configName + ", data: " + JSON.stringify(data, null, 4) + ", dirs: " + JSON.stringify(dirs, null, 4));
    var tableArray = tablePath.split('/');
    var collection = tableArray.shift();
    var entity = tableArray.shift();
    var fileName = GetFilePath(configName, collection, entity, dirs, false);
    if (!fs.existsSync(fileName)) { throw { name: "FileNotFoundException", message: "The requested file: \"" + fileName + "\" was not found" }; }
    console.log("Reading from file " + fileName);
    var jsonFile = JSON.parse("" + fs.readFileSync(fileName));
    var oldFile = jsonFile.document.converted.guidata;

    var curName = tableArray.shift();
    var dataIdx = Utils.ContainsName(oldFile, curName, "name");
    if (dataIdx < 0) {
        console.log("Cannot find data in file!");
        return;
    }
    var oldData = oldFile[dataIdx];
    console.log("oldData: " + JSON.stringify(oldData, null, 4));
    ReplaceByPath(oldData, tableArray, data);

    jsonFile.document.converted.guidata[dataIdx] = oldData;
    console.log("After replacement, table data is " + JSON.stringify(oldData, null, 4));

    var filePath = GetTempFilePath(configName, collection, entity, dirs);
    console.log("Writing to file " + filePath);
    fs.writeFileSync(filePath, JSON.stringify(jsonFile, null, 4));
};

/**
 * Removes a user's temporary directory, not saving changes to the database
 * @param {Object} dirs - Object containing paths to important directories
 * @param {string} dirs.tmp - The temporary directory for this editing session
 * @param {string} dirs.db - The database temporary storage directory
 * @param {string} dirs.trash - The directory to hold discarded configuration changes
 */
function DiscardWorkingDir(dirs) {
    console.log("Deleting existing trash dir (if any)");
    Utils.ExecSync("rm -rf " + dirs.trash);
    console.log("DiscardWorkingDir: Moving db temp to TRASH: mv " + dirs.db + " " + dirs.trash);
    Utils.ExecSync("mv " + dirs.db + " " + dirs.trash);
    console.log("DiscardWorkingDir: Moving temp files to TRASH: mv " + dirs.tmp + "/* " + dirs.trash);
    Utils.ExecSync("mv " + dirs.tmp + "/* " + dirs.trash);
    console.log("DiscardWorkingDir: Deleting temp directory: rmdir " + dirs.tmp);
    Utils.ExecSync("rmdir " + dirs.tmp);
};

/**
 * Creates a new configuration containing all modifications to the old configuration plus any unmodified files.
 * @param {string} oldConfig - Name of the old configuration
 * @param {string} newConfig - Name of the new configuration
 * @param {string[]} files - List of modified files
 * @param {Object} dirs - Object containing paths to important directories
 * @param {string} dirs.tmp - The temporary directory for this editing session
 * @param {string} dirs.db - The database temporary storage directory
 * @param {string} dirs.trash - The directory to hold discarded configuration changes
 * @throws DBOperationFailedException: More information in ex.message
 * @throws CollectionNameNotFound - Unable to determine collection name
 */
function SaveConfigurationChanges(oldConfig, newConfig, files, dirs, dbConfig) {
    console.log("Saving Configuration Changes, oldConfig: " + oldConfig + ", newConfig: " + newConfig + ", files: " + JSON.stringify(files, null, 4) + ", dirs: " + JSON.stringify(dirs, null, 4));
    var fileInfo = Connector.RunBuildFilterQuery(dbConfig, oldConfig).search;

    var f;
    for (f in files) {
        if (files.hasOwnProperty(f)) {
            console.log("Current file information: " + JSON.stringify(files[f], null, 4));
            var collectionName = files[f].collection;
            var entities = files[f].entities;
            var entity = files[f].entity ? files[f].entity : "notprovided";
            var version = files[f].version;
            var thisFileInfo = {};
            for (var fi in fileInfo) {
                if (fileInfo.hasOwnProperty(fi)) {
                    if (collectionName === fileInfo[fi].query.collection && Utils.ContainsString(entities, fileInfo[fi].query.filter["entities.name"]) != -1) {
                        console.log("Matched file information to Document: " + JSON.stringify(fileInfo[fi], null, 4));
                        thisFileInfo = fileInfo[fi];
                        entity = fileInfo[fi].query.filter["entities.name"];
                        fileInfo.splice(fi, 1);
                    }
                }
            }

            console.log("Getting metadata from original and changed files");
            var modified = GetFilePath(oldConfig, collectionName, entity, dirs, false);
            var newMetadata = ReadFileMetadata(dirs, thisFileInfo, dbConfig);
            //if (newMetadata.version === version) {
            //    version = Utils.Uniquify(version);
            //}
            newMetadata.version = version;
            console.log("newMetadata: " + JSON.stringify(newMetadata, null, 4));

            console.log("Checking metadata version strings");
            while (VersionExists(entity, collectionName, newMetadata.version, dbConfig)) {
                console.log("Inferring new version string...");
                version = Utils.Uniquify(newMetadata.version);
                console.log("Changing version from " + newMetadata.version + " to " + version);
                newMetadata.version = version;
                console.log("OK");
            }
            console.log("Prepending changelog");
            newMetadata.changelog = files[f].changelog + newMetadata.changelog;

            console.log("Writing new metadata to file");
            if (WriteFileMetadata(newMetadata, modified)) {

                console.log("Running store query");
                var data = "" + fs.readFileSync(modified);
                console.log("Writing " + data + " to database");
                Connector.RunStoreQuery(dbConfig, data, collectionName, newMetadata.version, entity, "gui", newConfig);
            } else {
                console.log("ERROR: Could not find file " + modified);
                console.log("Please check if it exists.");
            }
        }
    }

    console.log("Running addconfig for unmodified files: " + JSON.stringify(fileInfo, null, 4));
    for (f in fileInfo) {
        if (fileInfo.hasOwnProperty(f)) {
            var unmodifiedVersion = GetVersion(fileInfo[f], dirs, dbConfig);
            Connector.RunAddConfigQuery(dbConfig, newConfig, unmodifiedVersion, fileInfo[f].query.collection, { name: fileInfo[f].query.filter["entities.name"] });
        }
    }

    DiscardWorkingDir(dirs);
};

/**
 * Creates a new configuration using the provided array of entity information
 * @param {string} configName - The name of the new configuration
 * @param {Object} configData - An object holding the array of data
 * @param {Object[]} configData.entities - The list of entity information, consisting of name, version, and collection
 * @param {Object} dirs - Paths to filesystem directories for temporary storage
 * @throws DBOperationFailedException: More information in ex.message
 */
function CreateNewConfiguration(configName, configData, dbConfig) {
    console.log("CreateNewConfigration: configName: " + configName + ", configData: " + JSON.stringify(configData, null, 4));
    var query = {
        operations: []
    };

    for (var d in configData.entities) {
        if (configData.entities.hasOwnProperty(d)) {
            var entityData = configData.entities[d];
            console.log("Entity: " + JSON.stringify(entityData, null, 4));
            query.operations.push({
                filter: { version: entityData.version, "entities.name": entityData.name },
                configuration: configName,
                collection: entityData.collection,
                dbprovider: dbConfig.dbprovider,
                operation: "addconfig",
                dataformat: "gui"
            });
        }
    }

    Connector.RunNewConfigQuery(dbConfig, query);
};

/**
 * Reads any configuration-level metadata. Currently limited to a list of entity names and versions present ({}.entities)
 * @param {string} configPath - Name of the configuration
 * @returns {Object} Configuration metadata object
 * @throws DBOperationFailedException: More information in ex.message
 */
function ReadConfigurationMetadata(configName, dirs, dbConfig) {
    console.log("ReadConfigurationMetadata: configname: " + configName + ", dirs: " + JSON.stringify(dirs, null, 4));

    var data = Connector.RunBuildFilterQuery(dbConfig, configName ).search;

    var metadata = {
        entities: []
    };
    for (var i in data) {
        if (data.hasOwnProperty(i)) {
            console.log("Loading metadata: File " + i + " of " + data.length);
            var version = GetVersion(data[i], dirs, dbConfig);
            metadata.entities.push({ name: data[i].query.filter["entities.name"], file: data[i].name + "_" + data[i].query.collection, version: version, collection: data[i].query.collection });
        }
    }

    console.log("Returning entity list: " + JSON.stringify(metadata.entities, null, 4));
    return metadata;
};

/**
 * Gets the version of a given entity/collection in configName
 * @param {string} query - BuildFilter query 
 * @param {string} collection - Name of the collection
 * @param {string} configName - Name of the configuration
 * @returns {string} Version identifier
 * @throws DBOperationFailedException: More information in ex.message
 */
function GetVersion(query, dirs, dbConfig) {
    console.log("\"GetVersion\": {\"query\":" + JSON.stringify(query, null, 4) + ", \"dirs\":" + JSON.stringify(dirs, null, 4) + "},");
    var ver = LoadFile(query, dirs, dbConfig).data.version;
    console.log("GetVersion Returning " + ver);
    return ver;
};

/**
 * Reads the metadata objects from the given file
 * @param {Object} dirs - Object containing paths to important directories
 * @param {string} dirs.tmp - The temporary directory for this editing session
 * @param {string} dirs.db - The database temporary storage directory
 * @param {Object} query - Database query for file
 * @returns {Object} The metadata object 
 * @throws DBOperationFailedException: More information in ex.message
 */
function ReadFileMetadata(dirs, query, dbConfig) {
    console.log("ReadFileMetadata: query=" + JSON.stringify(query, null, 4) + ", dirs=" + JSON.stringify(dirs, null, 4) + ", dbConfig: " + JSON.stringify(dbConfig, null, 4));

    var jsonFile = LoadFile(query, dirs, dbConfig).data;
    if (jsonFile.changelog === undefined) {
        jsonFile.changelog = "";
    }
    var metadata = {
        entities: jsonFile.entities,
        bookkeeping: jsonFile.bookkeeping,
        aliases: jsonFile.aliases,
        configurations: jsonFile.configurations,
        version: jsonFile.version,
        changelog: jsonFile.changelog,
        collection: query.query.collection
    };

    console.log("ReadFileMetadata returning: " + JSON.stringify(metadata, null, 4));
    return metadata;
};

/**
 * Writes metadata objects to a file (for temporary storage)
 * @param {Object} newMetadata - Metadata object
 * @param {string} filebase - Display name of the configuration file to read
 * @returns {Boolean} - True if succeeded
 */
function WriteFileMetadata(newMetadata, fileName) {
    console.log("WriteFileMetadata: newMetadata=" + JSON.stringify(newMetadata, null, 4) + ", fileName=" + fileName);

    console.log("Reading file: " + fileName);
    if (!fs.existsSync(fileName)) return false;
    var jsonFile = JSON.parse("" + fs.readFileSync(fileName));


    console.log("Setting fields: " + JSON.stringify(newMetadata, null, 4));
    jsonFile.configurable_entity = newMetadata.configurable_entity;
    jsonFile.bookkeeping = newMetadata.bookkeeping;
    jsonFile.aliases = newMetadata.aliases;
    jsonFile.configurations = newMetadata.configurations;
    jsonFile.version = newMetadata.version;
    jsonFile.document.converted.changelog = newMetadata.changelog;

    console.log("Writing data to file");
    //console.log("fileName: " + fileName + ", metadata: " + JSON.stringify(jsonFile,null,4));
    fs.writeFileSync(fileName, JSON.stringify(jsonFile, null, 4));

    return true;
};

/**
 * Makes the appropriate directories for a given userId
 * @param {string} userId - Unique user identifier
 * @returns {Object} Directories object 
 */
function GetDirectories(userId, dbConfig) {
    console.log("GetDirectories: userid=" + userId);
    if (dbConfig.baseDir === "" || !dbConfig.baseDir) {
        dbConfig.baseDir = process.env["HOME"] + "/databases";
        console.log("WARNING: ARTDAQ_DATABASE_DATADIR not set. Using $HOME/databases instead!!!");
    }

    if (!fs.existsSync(dbConfig.baseDir)) {
        console.log("ERROR: Base Directory " +dbConfig.baseDir + " doesn't exist!!!");
        throw { name: "BaseDirectoryMissingException", message: "ERROR: Base Directory doesn't exist!!!" };
    }
    if (!fs.existsSync(path_module.join(dbConfig.baseDir, "db"))) {
        fs.mkdirSync(path_module.join(dbConfig.baseDir, "db"));
    }
    if (!fs.existsSync(path_module.join(dbConfig.baseDir, "tmp"))) {
        fs.mkdirSync(path_module.join(dbConfig.baseDir, "tmp"));
    }
    if (!fs.existsSync(path_module.join(dbConfig.baseDir, "TRASH"))) {
        fs.mkdirSync(path_module.join(dbConfig.baseDir, "TRASH"));
    }

    // ReSharper disable UseOfImplicitGlobalInFunctionScope
    var db = path_module.join(dbConfig.baseDir, "db", userId);
    var tmp = path_module.join(dbConfig.baseDir, "tmp", userId);
    var trash = path_module.join(dbConfig.baseDir, "TRASH", userId);
    // ReSharper restore UseOfImplicitGlobalInFunctionScope

    if (!fs.existsSync(db)) {
        fs.mkdirSync(db);
    }
    if (!fs.existsSync(tmp)) {
        fs.mkdirSync(tmp);
    }
    if (!fs.existsSync(trash)) {
        fs.mkdirSync(trash);
    }

    return { db: db, tmp: tmp, trash: trash };
};

/**
 * Check if a given version exists for a given entity-collection
 * @param {string} entity - Name of the Configurable Entity
 * @param {string} collection - Collection name
 * @param {string} version - Version identifier
 * @returns {Boolean} Whether the version already exists 
 * @throws DBOperationFailedException: More information in ex.message
 */
function VersionExists(entity, collection, version, dbConfig) {
    console.log("\"VersionExists\": { version:\"" + version + "\", entity:" + JSON.stringify(entity, null, 4) + ", collection: \"" + collection + "\"}");
    var query = {
        filter: {
            "entities.name": entity
        },
        collection: collection,
        dbprovider: dbConfig.dbprovider,
        operation: "findversions",
        dataformat: "gui"
    };
    var vers = Connector.RunGetVersionsQuery(dbConfig, query).search;
    console.log("Search returned: " + JSON.stringify(vers, null, 4));
    return Utils.ContainsName(vers, version, "name") >= 0;
};

/**
 * Either takes the database lock or returns false
 * @returns {Boolean} If the caller has the lock
 */
function Lock() {
    console.log("Lock");
    if (fs.existsSync("/tmp/node_db_lockfile")) {
        if (Date.now() - fs.fstatSync("/tmp/node_db_lockfile").ctime.getTime() > 1000) {
            console.log("Stale Lockfile detected, deleting...");
            return Unlock();
        } else {
            console.log("Lockfile detected and is not stale, aborting...");
        }
        return false;
    }

    fs.writeFileSync("/tmp/node_db_lockfile", "locked");
    return true;
}

/**
 * Releases the database lock
 * @returns {Boolean}  True when complete
 */
function Unlock() {
    console.log("Unlock");
    fs.unlinkSync("/tmp/node_db_lockfile");
    return true;
}

// POST calls
db.RO_GetData = function (post, dbConfig) {
    console.log("RO_GetData: " + JSON.stringify(post, null, 4));
    var ret = { Success: false, data: {} };
    try {
        ret.data = GetData(post.configName, post.collection, post.entity, GetDirectories(post.user, dbConfig));
        ret.Success = true;
    } catch (e) {
        console.log("Exception occurred: " + e.name + ": " + e.message);
    }

    console.log("GetData complete");
    return ret;
};

db.RW_MakeNewConfig = function (post, dbConfig) {
    if (Lock()) {
        console.log("RW_MakeNewConfig: Request to make new configuration received: " + JSON.stringify(post, null, 4));
        var res = { Success: false };
        var error = false;
        var e;
        try {
            var configs = Connector.RunGetConfigsQuery(dbConfig).search;
            if (!Utils.ValidatePath(post.name)) {
                console.log("Invalid name detected!");
                error = true;
            }
            while (Utils.ContainsName(configs, post.name, "name") >= 0) {
                console.log("Inferring new configuration name");
                post.name = Utils.Uniquify(post.name);
            }
        } catch (e) {
            error = true;
            console.log("Exception occurred: " + e.name + ": " + e.message);
        }
        if (!error) {
            console.log("Creating Configuration");
            try {
                CreateNewConfiguration(post.name, JSON.parse(post.config), dbConfig);
                res.Success = true;
            } catch (e) {
                console.log("Exception occurred: " + e.name + ": " + e.message);
            }
        }
        Unlock();
        console.log("MakeNewConfig completed");
        return res;
    }
    return null;
};

db.RW_saveConfig = function (post, dbConfig) {
    if (Lock()) {
        console.log("RW_saveConfig: Request to save configuration recieved. Configuration data: " + JSON.stringify(post, null, 4));
        var res = { Success: false };
        var error = false;
        var e;
        try {
            console.log("Checking for unique Configuration name");
            var configs = Connector.RunGetConfigsQuery(dbConfig).search;
            if (!Utils.ValidatePath(post.newConfigName) || Utils.ContainsName(configs, post.oldConfigName, "name") < 0) {
                console.log("Invalid name detected!");
                error = true;
            }
            while (Utils.ContainsName(configs, post.newConfigName, "name") >= 0) {
                console.log("Inferring new configuration name");
                post.newConfigName = Utils.Uniquify(post.newConfigName);
            }
        } catch (e) {
            error = true;
            console.log("Exception occurred: " + e.name + ": " + e.message);
        }
        if (!error) {
            try {
                console.log("Updating Configuration Files");
                SaveConfigurationChanges(post.oldConfigName, post.newConfigName, post.files, GetDirectories(post.user, dbConfig), dbConfig);
                res.Success = true;
            } catch (e) {
                console.log("Exception occurred: " + e.name + ": " + e.message);
            }
        }
        Unlock();
        console.log("SaveConfig completed");
        return res;
    }
    return null;
};

db.RO_LoadNamedConfig = function (post, dbConfig) {
    console.log("RO_LoadNamedConfig: Request for configuration with name \"" + post.configName + "\" and search query \"" + post.query + "\" received.");
    if (post.query.length === 0 || post.configName === "No Configurations Found") {
        return { collections: [] };
    }
    return LoadConfigFiles(post.configName, GetDirectories(post.user, dbConfig), JSON.parse(post.query), dbConfig);
};

db.RW_discardConfig = function (post, dbConfig) {
    console.log("RW_discardConfig: Discarding configuration with parameters: " + JSON.stringify(post, null, 4));
    DiscardWorkingDir(GetDirectories(post.user, dbConfig));
    return { Success: true };
};

db.RO_AddOrUpdate = function (post, dbConfig) {
    console.log("RO_AddOrUpdate: Request to update table row recieved: " + JSON.stringify(post, null, 4));
    UpdateTable(post.configName, post.table, post.row, GetDirectories(post.user, dbConfig));
    return { Success: true };
}

db.RO_Update = function (post, dbConfig) {
    console.log("RO_Update: Request to update table received: " + JSON.stringify(post, null, 4));
    UpdateTable(post.configName, post.table, { id: post.id, name: post.name, column: post.column, value: post.value }, GetDirectories(post.user, dbConfig));
    console.log("Update Complete");
    return { Success: true };
};

db.RO_LoadConfigMetadata = function (post, dbConfig) {
    console.log("RO_LoadConfigMetadata: Request to load configuration metadata received: " + JSON.stringify(post, null, 4) + ", dbConfig: " + JSON.stringify(dbConfig, null, 4));
    var ret = { Success: false, data: {} };
    try {
        ret.data = ReadConfigurationMetadata(post.configName, GetDirectories(post.user, dbConfig), dbConfig);
        ret.Success = true;
    } catch (e) {
        console.log("Exception caught: " + e.name + ": " + e.message);
    }
    console.log("LoadConfigMetadata complete");
    return ret;
};

db.RO_LoadFileMetadata = function (post, dbConfig) {
    console.log("RO_LoadFileMetadata: Request to load file metadata received: " + JSON.stringify(post, null, 4));
    var ret = { Success: false, data: {} };
    var dirs = GetDirectories(post.user, dbConfig);
    var error = false;
    var query = {};
    var e;
    try {
        var search = Connector.RunBuildFilterQuery(dbConfig, post.configName).search;
        for (var s in search) {
            if (search.hasOwnProperty(s)) {
                if (search[s].query.collection === post.collection && search[s].query.filter["entities.name"] === post.entity) {
                    query = search[s];
                }
            }
        }
    } catch (e) {
        error = true;
        console.log("Exception caught: " + e.name + ": " + e.message);
    }
    if (!error) {
        try {
            ret.data = ReadFileMetadata(dirs, query, dbConfig);
            ret.Success = true;
        } catch (e) {
            console.log("Exception caught: " + e.name + ": " + e.message);
        }
    }
    console.log("LoadFileMetadata complete");
    return ret;
};

db.RW_UploadConfigurationFile = function (post, dbConfig) {
    if (Lock()) {
        console.log("RW_UploadConfigurationFile: Recieved request to upload file: " + JSON.stringify(post, null, 4));
        var e;
        var error = false;
        var ret = { Success: false };
        try {
            while (VersionExists(post.entity, post.collection, post.version, dbConfig)) {
                console.log("Version already exists. Running uniquifier...");
                post.version = Utils.Uniquify(post.version);
            }
        } catch (e) {
            error = true;
            console.log("Exception caught: " + e.name + ": " + e.message);
        }

        if (!error) {
            console.log("Running store fhicl query");
            try {
                Connector.RunStoreQuery(dbConfig, post.file, post.collection, post.version, post.entity, post.type);
                ret.Success = true;
            } catch (e) {
                console.log("Exception caught: " + e.name + ": " + e.message);
            }
        }
        Unlock();
        console.log("UploadConfigurationFile complete");
        return ret;
    }
    return null;
};

db.RO_DownloadConfigurationFile = function (post, dbConfig) {
    console.log("RO_DownloadConfigurationFile: Request to download file(s) received: " + JSON.stringify(post, null, 4));
    var dirs = GetDirectories(post.user, dbConfig);
    var configObj = JSON.parse(post.config);
    try {
        if (configObj.entities.length === 1) {
            console.log("Single file mode: Fetching file...");
            var fileInfo = FetchFile(configObj.entities[0], post.type, dirs.db, dbConfig);

            var fclhdrs = {
                'Content-Type': 'text/plain',
                'Content-Length': fileInfo.size,
                'Content-Disposition': 'attachment filename=' + fileInfo.fileName
            }
            console.log("Headers: " + JSON.stringify(fclhdrs, null, 4) + ", fileInfo: " + JSON.stringify(fileInfo, null, 4));
            
            var fclStream = fs.createReadStream(fileInfo.filePath);
            db.emit("stream", fclStream, fclhdrs, 200);
        } else if (configObj.entities.length > 1) {
            var args = ['cz'];
            for (var e in configObj.entities) {
                if (configObj.entities.hasOwnProperty(e)) {
                    args.push(FetchFile(configObj.entities[e], post.type, dirs.db, dbConfig).fileName);
                }
            }
            var fileName = post.tarFileName + ".tar.gz";
            var tarhdrs = {
                'Content-Type': "application/x-gzip",
                'Content-Disposition': 'attachment filename=' + fileName
            }

            console.log("Spawning: tar " + args.join(" "));
            var tar = child_process.spawn("tar", args, { cwd: dirs.db, stdio: [0, 'pipe', 0] });
            db.emit("stream", tar.stdout, tarhdrs, 200);

        }
    } catch (err) {
        console.log("Exception caught: " + err.name + ": " + err.message);

        var s = new stream.Readable();
        s._read = function noop() { };
        s.push("ERROR");
        s.push(null);

        var errhdrs = {
            'Content-Type': 'text/plain'
        }
        db.emit("stream", s, errhdrs, 500);
    }
    //Stream emit has its own 'end', no return value necessary
};

db.RO_NamedConfigs = function (post, dbConfig) {
    console.log("RO_NamedConfigs: Request for Named Configurations received");
    var filter = "" + post.configFilter;
    return FindConfigs(filter, dbConfig);
};

db.RW_updateDbConfig = function (post, dbConfig) {
    console.log("RW_updateDbConfig: Request to update module configuration received: " + JSON.stringify(post, null, 4));
    var output = { Success: false };
    if (fs.existsSync(post.baseDir) && (post.dbprovider === "filesystem" || post.dbprovider === "mongo")) {
        if (dbConfig.baseDir !== post.baseDir) {
            dbConfig.baseDir = post.baseDir;
            db.emit("message", { name: "db", target: "baseDir", data: post.baseDir });
        }
        if (dbConfig.dbprovider !== post.dbprovider) {
            dbConfig.dbprovider = post.dbprovider;
            db.emit("message", { name: "db", target: "dbprovider", data: post.dbprovider });
        }
        if (dbConfig.instanceName !== post.instanceName) {
            dbConfig.instanceName = post.instanceName;
            db.emit("message", { name: "db", target: "instanceName", data: post.instanceName });
        }
        console.log("DB Config is now: " + JSON.stringify(dbConfig, null, 4));
        output.Success = true;
    }
    return output;
}

db.RO_SearchLoadedConfig = function (post, dbConfig) {
    console.log("RO_SearchLoadedConfig: Searching for keys containing name \"" + post.searchKey + "\" from configuration name \"" + post.configName + "\" received.");
    if (post.configName === "No Configurations Found") {
        return { Success: false, collections: [] };
    }
    return SearchConfigFiles(post.searchKey, post.configName, GetDirectories(post.user, dbConfig), dbConfig);
}

db.RW_makeNewDBInstance = function (post, dbConfig) {
    console.log("RW_makeNewDBInstance: Creating new database instance \"" + post.name + "\".");
    dbConfig.instanceName = post.name;
    return { Success: true };
}

db.RW_AddEntityToFile = function (post, dbConfig) {
    console.log("RW_AddEntityToFile: request to add entity name \"" + post.name + "\" to configuration file " + post.configName + "/" + post.collection + "/" + post.entity); 
    return { Success: true };
}

// GET calls
db.GET_EntitiesAndVersions = function (dbConfig) {
    console.log("GET_EntitiesAndVersions: Request for current Entities and Versions received");
    var output = {
        Success: false,
        collections: []
    };
    try {
        var entities = Connector.RunGetEntitiesQuery(dbConfig).search;
        console.log("Returned entities: " + JSON.stringify(entities, null, 4));
        for (var ent in entities) {
            if (entities.hasOwnProperty(ent)) {
                var entity = entities[ent];
                var versions = Connector.RunGetVersionsQuery(dbConfig, entity.query);
                if (Utils.ContainsName(output.collections, entity.query.collection, "name") < 0) {
                    output.collections.push({ name: entity.query.collection, entities: [] });
                }

                var index = Utils.ContainsName(output.collections, entity.query.collection, "name");

                var entityObj = {
                    collection: entity.query.collection,
                    name: entity.name,
                    versions: versions
                };
                output.collections[index].entities.push(entityObj);
            }
        }
        output.Success = true;
    } catch (e) {
        console.log("Exception caught: " + e.name + ": " + e.message);
    }
    console.log("EntitiesAndVersions complete");
    return output;
};

db.GET_getDbConfig = function (dbConfig) {
    console.log("GET_getDbConfig Request for Database Module Configuration received");

    var instances = FindInstances(dbConfig);
    var output = {
        Success: true,
        baseDir: dbConfig.baseDir,
        dbprovider: dbConfig.dbprovider,
        instanceName: dbConfig.instanceName,
        data: instances.data
    };
    return output;
}

// Serverbase Module definition
db.MasterInitFunction = function (workerData, config) {
    var dbConfig = MakeDbConfig(config);
    workerData["db"] = dbConfig;
    GetDirectories("", dbConfig);
};

module.exports = function (moduleHolder) {
    moduleHolder["db"] = db;
};
